/*
 * MakeINES v1.0, by Quietust
 *
 * Makefile structure
 *	MAKEINES	- file identification tag - MUST be the first line in the file
 *	filename.nes	- Name of the .NES file to generate
 *	MAPR nn		- Specify mapper number
 *	4SCR n		- Specify whether or not there is 4-screen VRAM
 *	BATT n		- Specify whether or not there is a battery
 *	VMIR n		- Choose between vertical and horizontal mirroring
 *	PROG filename	- Name of PRG ROM file
 *	CHAR filename	- Name of CHR ROM file (optional)
 */

#include <stdio.h>

#define	MAXLEN	256

char path[MAXLEN];
void	SetPath (const char *filename)
{
	int i;
	strcpy(path,filename);
	for (i = strlen(path) - 1; i >= 0; i--)
		if (path[i] == '\\')
			break;
	path[i+1] = 0;
}

char fname[MAXLEN];
char *	GetPath (const char *filename)
{
	strcpy(fname,path);
	strcat(fname,filename);
	return fname;
}

int	SplitLine (const char *line, char *header, char *data)
{
	int i;
	int j = 0;
	int hoff = 0, doff = 0;
	for (i = 0; i < strlen(line); i++)
	{
		if (j == 0)
		{
			if (line[i] == ' ')
				j = 1;
			else	header[hoff++] = line[i];
		}
		else	data[doff++] = line[i];
	}
	header[hoff] = 0;
	data[doff] = 0;
	if (header[hoff-1] == '\n')
		return 0;
	if (data[doff-1] == '\n')
	{
		data[doff-1] = 0;
		return 1;
	}
	else	return 0;
}

int main (int argc, char **argv)
{
	FILE *makeinfo;
	FILE *output;
	FILE *prg = NULL, *chr = NULL;
	unsigned char prgs = 0, chrs = 0, flg0 = 0, flg1 = 0, x;
	char data[MAXLEN];
	int i;
	if (argc < 2)
	{
		fprintf(stderr,"Syntax: %s <makefile>\n",argv[0]);
		return 1;
	}
	makeinfo = fopen(argv[1],"rt");
	if (makeinfo == NULL)
	{
		fprintf(stderr,"Unable to open makefile!\n");
		return 1;
	}
	SetPath(argv[1]);
	fgets(data,MAXLEN,makeinfo);
	if (strcmp(data,"MAKEINES\n"))
	{
		fclose(makeinfo);
		fprintf(stderr,"Specified file is not a valid iNES makefile!\n");
		return 1;
	}
	fgets(data,MAXLEN,makeinfo);
	if (data[strlen(data)-1] != '\n')
	{
		fclose(makeinfo);
		fprintf(stderr,"Filename too long! Increase MAXLEN and recompile.\n");
		return 1;
	}
	else	data[strlen(data)-1] = 0;
	output = fopen(GetPath(data),"wb");
	if (output == NULL)
	{
		fclose(makeinfo);
		fprintf(stderr,"Unable to open file %s for output!\n",data);
		return 1;
	}
	printf("Opened file %s for output.\n",data);
	while (fgets(data,MAXLEN,makeinfo))
	{
		char blk[MAXLEN];
		char args[MAXLEN];
		if (feof(makeinfo))
			break;
		if (!SplitLine(data,blk,args))
		{
			fclose(makeinfo);
			fclose(output);
			fprintf(stderr,"Error decoding line: %s\n",data);
			return 1;
		}
		if (!memcmp(blk,"MAPR",4))
		{
			sscanf(args,"%d",&i);
			flg0 |= (i & 0x0F) << 4;
			flg1 |= (i & 0xF0);
			printf("Using mapper number %i\n",i);
		}
		else if (!memcmp(blk,"4SCR",4))
		{
			sscanf(args,"%d",&i);
			if (i)
			{
				flg0 |= 0x08;
				printf("Enabling 4-screen VRAM\n");
			}
		}
		else if (!memcmp(blk,"BATT",4))
		{
			sscanf(args,"%d",&i);
			if (i)
			{
				flg0 |= 0x02;
				printf("Enabling battery-backed SRAM\n");
			}
		}
		else if (!memcmp(blk,"VMIR",4))
		{
			sscanf(args,"%d",&i);
			if (i)
			{
				flg0 |= 0x01;
				printf("Using vertical mirroring\n");
			}
			else	printf("Using horizontal mirroring\n");
		}
		else if (!memcmp(blk,"VSUS",4))
		{
			sscanf(args,"%d",&i);
			if (i)
			{
				flg1 |= 0x01;
				printf("Setting game as VS Unisystem ROM\n");
			}
		}
		else if ((!memcmp(blk,"PROG",4)) || (!memcmp(blk,"CHAR",4)))
		{
			FILE *rom = fopen(GetPath(args),"rb");
			int blklen;
			if (rom == NULL)
			{
				fprintf(stderr,"Error reading %s data - unable to open file '%s'!\n",blk,args);
				continue;
			}
			fseek(rom,0,SEEK_END);
			blklen = ftell(rom);
			fseek(rom,0,SEEK_SET);
			if (!memcmp(blk,"PROG",4))
			{
				prg = rom;
				prgs = blklen / 0x4000;
				if (blklen & 0x3FFF)
				{
					if (prg)	fclose(prg);
					if (chr)	fclose(chr);
					fclose(makeinfo);
					fprintf(stderr,"Error reading PRG data - size (0x%X) is not a multiple of 16KB!\n",blklen);
					return 1;
				}
				printf("Storing %i blocks (%i bytes) of PRG data\n",prgs,blklen);
			}
			if (!memcmp(blk,"CHAR",4))
			{
				chr = rom;
				chrs = blklen / 0x2000;
				if (blklen & 0x1FFF)
				{
					if (prg)	fclose(prg);
					if (chr)	fclose(chr);
					fclose(makeinfo);
					fprintf(stderr,"Error reading CHR data - size (0x%X) is not a multiple of 8KB!\n",blklen);
					return 1;
				}
				printf("Storing %i blocks (%i bytes) of CHR data\n",chrs,blklen);
			}
		}
		else	fprintf(stderr,"Unrecognized directive %s encountered, skipping...\n",blk);
	}
	fclose(makeinfo);
	printf("Generating .NES file...");
	fwrite("NES\x1A",4,1,output);
	fwrite(&prgs,1,1,output);
	fwrite(&chrs,1,1,output);
	fwrite(&flg0,1,1,output);
	fwrite(&flg1,1,1,output);
	x = 0;
	for (i = 0; i < 8; i++)
		fwrite(&x,1,1,output);
	for (i = 0; i < prgs * 16384; i++)
	{
		fread(&x,1,1,prg);
		fwrite(&x,1,1,output);
	}
	for (i = 0; i < chrs * 8192; i++)
	{
		fread(&x,1,1,chr);
		fwrite(&x,1,1,output);
	}
	fclose(output);
	printf("done!\n");
	return 0;
}